import cv2


def video_resolution(video_path):
    """
    Get the resolution of the video file

    This function returns a string which is the resolution of the video file fed

    Parameters
    -----------
    video_path : path to the specific video

    Returns
    -------
    string
        '1680'
    """

    vid = cv2.VideoCapture(video_path)
    height = vid.get(cv2.CAP_PROP_FRAME_HEIGHT)
    resolution = int(height)
    return resolution



import audioop
import wave


def rms(wav_path):
    """
    RMS level (root mean squared) is just proportional to the amount of energy over a period of time in the signal. This can be used to distinguish audios that are louder from each other.


    This function returns the rms value of a given function.

    Parameters
    -----------
    wav_path : path to a .wav audio

    Returns
    -------
    int: the rms value of the audio
        "880"
    """
    wav = wave.open(wav_path)
    frames, width = wav.readframes(wav.getnframes()), wav.getsampwidth()
    aud_rms = audioop.rms(frames, width)
    return aud_rms



import cv2
import torch


def load_model(modelname):
    """
    Loads the object detection model from the torch hub
    Parameters
    ----------
    None
    Returns
    -------
    callable, list
        model, ['person', 'bicycle', ... 'hair brush']
    """

    model = torch.hub.load('ultralytics/yolov5',
                            modelname)
    classes = model.names
    
    return model, classes


def detect_objects(video_file, model, classes):
    """
    Get the objects present in the video

    This function returns the a set of objects found in the video file

    Parameters
    -----------
    video_file : path to a specific video 
    model : CNN model used for object detection
    classes: labels of the coco dataset

    Returns
    -------
    set
        {'person', 'cell phone', 'chair', 'tv'}
    """

    # model, classes = load_model(modelname)

    cap = cv2.VideoCapture(video_file)
    objects = list()
    while True:
        ret, frame = cap.read()
        if ret:
            result = model(frame)
            result.show()
            labels = result.xyxyn[0][:, -1].cpu().numpy()
            for index in range(len(labels)):
                objects.append(classes[int(labels[index])])
        else:
            return list(set(objects))
            


import os

from moviepy.editor import *


def bitrate(video_path, audio_path):
    """
    Get the bitrate of the video file

    This function returns the bit rate of the video file

    Parameters
    -----------
    video_path : path to the specific video
    audio_path : path to the specific audio

    Returns
    -------
    int
        1240
    """

    video = VideoFileClip(video_path)
    mp3_size = os.path.getsize(audio_path)
    vid_size = os.path.getsize(video_path)
    duration = video.duration
    bit_rate = int((((vid_size - mp3_size)/duration)/1024*8))

    return bit_rate


import numpy as np
import scipy.stats as stats
from collections import Counter
import math
from QPrism.Sensor.pipeline_functions.preprocessing import *


def sigmoid(x):
    return 1/(1 + np.exp(-x))


def consistency(data):
    data = np.asanyarray(data)
    co_var = abs(stats.variation(data, nan_policy='omit'))
    if math.isnan(co_var):
        co_var = 0
    consis = (1 - sigmoid(co_var))*2
    return consis
    #return 1/co_var if co_var != 0 else np.inf


def mode(vals):
    if (len(np.unique(vals))!=len(vals)):
        return Counter(vals).most_common(1)[0][0]
    elif (len(vals)<=1):
        return Counter(vals).most_common(1)[0][0]
    else:
        mean = np.mean(vals)
        std = np.std(vals)
        non_outlier_vals = []
        for val in vals:
            if (val < (mean+(3*std))) and ((val > (mean-(3*std)))):
                non_outlier_vals.append(val)
        return min(non_outlier_vals)


def valid_data_ratio(record, feature_names):
    record = features_to_float(record, feature_names)
    invalid_data_count = 0
    total_data_count = 0
    for feature in feature_names:
        invalid_data_count += record[feature].isna().sum()
        total_data_count += record[feature].shape[0]
    return (total_data_count-invalid_data_count)/total_data_count


def minmax(record, feature_names):
    record = features_to_float(record, feature_names)
    minmax_list = []
    for feature in feature_names:
        minmax_list.append(record[feature].max()-record[feature].min())
    return np.mean(minmax_list)
    #return np.mean([np.max([r[feature] for r in record]) - np.min([r[feature] for r in record]) for feature in feature_names])


def sampling_rate_consistency(record):
    #record = normalize_timestamp(record, ((list(record.keys()))[0]))
    sampling = [record.iloc[i+1][(list(record.keys()))[0]] - record.iloc[i][(list(record.keys()))[0]] for i in range(len(record) - 1)]
    sampling_trunc = [round(r, 3) for r in sampling]
    return consistency(sampling), np.median(sampling), mode(sampling_trunc)


def count_missing_segments(record, sample_mode):
    """missing_segment_count = 0
    inserts = []
    missing_amount = 0
    for i, r in enumerate(record[:-1]):
        if record[i + 1][(list(record[i+1].keys()))[0]] - record[i][(list(record[i].keys()))[0]] > sample_mode * 20:
            missing_segment_count += 1
            missing_amount += record[i + 1][(list(record[i+1].keys()))[0]] - record[i][(list(record[i].keys()))[0]]
            inserts.append(i)
    for count, i in enumerate(inserts):
        record = record[:i + 1 + count * 20] + [{'univariate_value': 0, 'inserted': True}] * 20 + record[i + 1 + count * 20:]

    return missing_amount, missing_segment_count, record"""
    missing_points = 0
    for i in range(record.shape[0]-1):
        if (record.iloc[i+1][(list(record.keys()))[0]] - record.iloc[i][(list(record.keys()))[0]]) >= (2*sample_mode):
            missing_points += (record.iloc[i+1][(list(record.keys()))[0]] - record.iloc[i][(list(record.keys()))[0]] - sample_mode) // sample_mode
    return missing_points/(record.shape[0]+missing_points)



import pandas as pd
import numpy as np


def get_feature_name_record_df(record_df):
    keys = list(record_df.keys())
    keys.remove(keys[0])
    return keys


def scale_data(data):
    scaled_data = (data - np.min(data)) / (np.max(data) - np.min(data))
    scaled_data = pd.DataFrame(scaled_data)
    scaled_data.columns = data.columns
    return scaled_data


import csv
import os

import pandas as pd
from pathlib import Path

# Video functions
from QPrism.Video.bit_rate import bitrate
from QPrism.Video.brightness import video_brightness
from QPrism.Video.creation_time import creation_time
from QPrism.Video.frame_rate import fps
from QPrism.Video.object_detection import detect_objects, load_model
from QPrism.Video.video_format import video_format
from QPrism.Video.video_length import video_length
from QPrism.Video.video_resolution import video_resolution
from QPrism.Video.artifacts import noise_detection

# Helpers 
from QPrism.Video.helpers.extract_audios import extract_audios
from QPrism.Video.helpers.video_paths import video_paths


class Video_DQM:

    def __init__(self):
        self.vid_metrics = ['bitrate', 'brightness', 'creation_time', 'framerate',\
                        'video_format', 'video_length', 'video_resolution']


    def duration(self, path:str):
        """
        This function gives the length for both a folder of audios or a single audio file.

        Returns the length of the input audio if the input path is a file. 
        Returns a dict matching the input audio files to their length if the input path is a folder.

        Parameters
        ----------
            path : str
                A path to a folder or a file.

        """

        if os.path.isdir(path):
            video_files = video_paths(path)
            length = {}
            
            for file in video_files:
                length[str(Path(file).stem)] = video_length(file)
            
            return pd.DataFrame(length.items(), columns=['Videos', 'Length'])

        else:
            return video_length(path)


    def resolution(self, path:str):
        """
        This function returns the resolution of a video or the videos in given folder

        Parameters
        ----------
            path : str
                path to a folder or a file
        """

        if os.path.isdir(path):
            video_files = video_paths(path)
            resolution = {}
            
            for file in video_files:
                resolution[str(Path(file).stem)] = video_resolution(file)
            
            return pd.DataFrame(resolution.items(), columns=['Videos', 'Resolution'])

        else:
            return video_resolution(path)


    def format(self, path:str):
        """
        Get the format of the video
        This functions calls the video_format function and 
        gives the format for both a folder of videos or a single video file.

        Parameters
        ----------
        path : path to a folder or a file

        Returns
        -------
        dict (incase of folder)
        str (incase of file)
            dict : {'video_file1' : '.mp4', ... 'video_filen' : '.mp4'}
            str : '.mp4'
        """

        if os.path.isdir(path):
            video_files = video_paths(path)
            format = {}
            
            for file in video_files:
                format[str(Path(file).stem)] = video_format(file)
            
            return pd.DataFrame(format.items(), columns=['Videos', 'Format'])

        else:
            return video_format(path)


    def bit_rate(self, path:str):
        """
        Get the bit rate of the video
        
        Returns the bit rate for input video or each video in the input folder.

        Parameters
        ----------
            path : str
                path to a folder or a file
        """

        if os.path.isdir(path):
            video_files = video_paths(path)
            bps = {}
            for file in video_files:
                video_name = str(Path(file).stem)
                audio_name = Path(video_name + "_Audio.mp3")
                audio_path = os.path.join(os.getcwd(), 'audio_files', audio_name)
                if not os.path.isfile(audio_path):
                    print("Audio file not found!!!")
                    print("Extracting audio from video")
                    extract_audios(file)
                bps[video_name] = bitrate(file, audio_path)

            return pd.DataFrame(bps.items(), columns=['Videos', 'Bitrate'])

        else:
            video_name = str(Path(path).stem)
            audio_name = Path(video_name + "_Audio.mp3")
            audio_path = os.path.join(os.getcwd(), 'audio_files', audio_name)
            if not os.path.isfile(audio_path):
                print("Audio file not found!!!")
                print("Extracting audio from video")
                extract_audios(path)

            return bitrate(path, audio_path)


    def object_detection(self, path:str, modelname:str):
        """
        This function detect all the objects present in the video.

        Returns the detected objects in the video if the input path is a file.
        Returns a dict matching the input videos to their detected objects.

        Parameters
        ----------
            path : str
                path to a folder or a file
        """

        if os.path.isdir(path):
            model, classes = load_model(modelname)
            video_files = video_paths(path)
            objects = {}
            
            for file in video_files:
                objects[str(Path(file).stem)] = detect_objects(file, model, classes)
            
            return pd.DataFrame(objects.items(), columns=['Videos', 'Objects'])

        else:
            model, classes = load_model(modelname)
            return detect_objects(path, model, classes)


    def framerate(self, path:str):
        """
        This function get the framerate of the video.

        Returns the frame rate of the video if the input path is a file.
        Returns a dict matching the input videos to their frame rate if the input path is a folder.

        Parameters
        ----------
            path : str
                path to a folder or a file
        """

        if os.path.isdir(path):
            video_files = video_paths(path)
            framerate = {}
            
            for file in video_files:
                framerate[str(Path(file).stem)] = fps(file)
            
            return pd.DataFrame(framerate.items(), columns=['Videos', 'Framerate'])

        else:
            return fps(path)


    def illumination(self, path:str):
        """
        This function gives the brightness of the video.

        Returns the brightness of the video if the input path is a file.
        Returns a dict matching the input videos to their frame rate if the input path is a folder.

        Parameters 
        ----------
            path : str
                path to a folder or a file
        """

        if os.path.isdir(path):
            video_files = video_paths(path)
            b = {}
            
            for file in video_files:
                b[str(Path(file).stem)] = video_brightness(file)
            
            return pd.DataFrame(b.items(), columns=['Videos', 'Brightness'])
        else:
            return video_brightness(path)


    def creation_time(self, path:str):
        """
        This function gets the creation time of a video.

        Returns the creation time of the video if the input path is a file.
        Returns a dict matching the input videos to their creation time if the input path is a folder.

        Parameters
        -----------
            path : str
                path to a folder or a file
        """

        if os.path.isdir(path):
            video_files = video_paths(path)
            time = {}
            
            for file in video_files:
                time[str(Path(file).stem)] = creation_time(file)
            
            return pd.DataFrame(time.items(), columns=['Videos', 'Date created'])

        else:
            return creation_time(path)


    def artifacts_ratio(self, path:str):
        """
        This function gets how much of the video that contains video artifacts

        Returns the ratio of the video that contains artifacts (motion blur, too grainy, static) if the input path is a file.
        Returns a dict mathing the input videos to their artifacts percentage if the input path is a folder.

        Parameters
        -----------
            path : str
                path to a folder or a file
        """

        if os.path.isdir(path):
            video_files = video_paths(path)
            percents = {}
            
            for file in video_files:
                percents[str(Path(file).stem)] = noise_detection(file)
            
            return pd.DataFrame(percents.items(), columns=['Videos', 'Artifacts'])
        else:
            return noise_detection(path)


    def save_csv(self, path:str, output_path:str, modelname='yolov5s'):
        """
        Creates a csv file consists of all the video metrics.

        Saves a csv file to the input path with all video quality metrics 

        Parameters
        ----------
            path : str
                path to single video file or folder
            output_path : str
                path to the output csv file
            modelname : str
                yolov5 model name for object detection
        """

        if os.path.isdir(path):
            video_files = video_paths(path)
        else:
            video_files = [os.path.abspath(path)]

        with open('video_metrics.csv', 'w') as file:
            header = ['Video Name', 'Bit Rate', 'Illumination', 'Date of Creation', 'Frame Rate', \
                        'Format', 'Duration', 'Resolution', 'Object Detection', 'Artifacts Ratio']
            writer = csv.writer(file)
            writer.writerow(header)

        csv_path = output_path
        metrics = pd.DataFrame()

        for i, video in enumerate(video_files):
            data_row = []

            data_row.append(str(Path(video).stem))
            data_row.append(self.bit_rate(video))
            data_row.append(self.illumination(video))
            data_row.append(self.creation_time(video))
            data_row.append(self.framerate(video))
            data_row.append(self.format(video))
            data_row.append(self.duration(video))
            data_row.append(self.resolution(video))
            data_row.append(self.object_detection(video, modelname))
            data_row.append(self.artifacts_ratio(video))

            data_series = pd.Series(data_row, index=header)
            metrics = metrics.append(data_series, ignore_index=True)
        
        print('Metrics saved to ', csv_path)
        metrics.to_csv(csv_path, index=False)


import numpy as np
import scipy.stats as stats

def classify_activities(record_set, record_len):
    n = len(record_set)
    Seg_Class = []
    for j in range(n):
        vals = [r['univariate_value'] for r in record_set[j] if 'inserted' not in r]
        if len(vals)>1:
            var = stats.variation(vals)
            if var < 0.01:
                Seg_Class.append(0)
            else:
                Seg_Class.append(1)
    for j in range(2, len(Seg_Class)):
        if j > 1:
            if (Seg_Class[j-1] and not (Seg_Class[j - 2]) and Seg_Class[j]) and (len(record_set[ j-1]) < record_len*.01):  # preceding seg is non-wear, following seg is non-wear, this seg is normal AND shorter than 1 sec --> 5 points
                Seg_Class[j-1] = 2

    return Seg_Class

def spurious_segment_density(Seg_Class):
    return len([s for s in Seg_Class if s == 2]) / len(Seg_Class)

def activity_density(Seg_Class):
    return len([s for s in Seg_Class if s == 1]) / len(Seg_Class)
def non_activity_segment_density(Seg_Class):
    return len([s for s in Seg_Class if s == 0]) / len(Seg_Class)


import matplotlib.pyplot as plt
import numpy as np
import wave
from pathlib import Path


def waveform(wav_path):
    """
    Get the waveform graph of an wav audio file.

    This function returns the time/signal waveform graph of a audio file.

    Parameters
    -----------
    wav_path : path to a .wav audio

    Returns
    -------
    Graph: wafeform graph
    """
    raw = wave.open(wav_path)
    signal = raw.readframes(-1)
    signal = np.frombuffer(signal, dtype="int16")
    f_rate = raw.getframerate()
    time = np.linspace(
        0,  # start
        len(signal) / f_rate,
        num=len(signal)
    )
    audio_name = str(Path(wav_path).stem)

    plt.figure(1)
    plt.title(audio_name + ' Sound Wave')
    plt.xlabel("Time")
    plt.plot(time, signal)
    plt.show()



import pandas as pd
import numpy as np
from QPrism.Sensor.data_processing import *
from QPrism.Sensor.pipeline_functions.single_function_DQMs import sampling_rate_consistency


def compute_SRC_single(record_df):
    sample_consistence, sample_rate_median, sample_mode = sampling_rate_consistency(record_df)
    return sample_consistence


def compute_SRC_multiple(data_list):
    SRC_list = []
    for record in data_list:
        record_SRC = compute_SRC_single(record)
        SRC_list.append(record_SRC)
    return np.mean(SRC_list)


import os
import pandas as pd
import json


def load_data_json_single_file(path):
    f = open(path)
    try:
        json_file = json.load(f)
    except:
        return pd.DataFrame([])
    json_file = json.dumps(list(json_file['records']))
    record_df = (pd.read_json(json_file))
    record_df[list(record_df.keys())[0]] = record_df[list(record_df.keys())[0]].apply(lambda x: x.value)
    return record_df


def load_data_json_multi_file(path):
    data_list = []
    records = os.listdir(path)
    for record in records:
        record_path = path+'/'+record
        try:
            data_list.append(load_data_json_single_file(record_path))
        except:
            data_list.append(pd.DataFrame())
    return data_list


def load_data_csv_one_file(path):
    record_df = pd.read_csv(path)
    keys = list(record_df.keys())
    record_df = record_df.drop(keys[0], axis=1)
    return record_df


def load_data_csv_multi_file(path):
    data_list = []
    records = os.listdir(path)
    for record in records:
        record_path = path+'/'+record
        try:
            data_list.append(load_data_csv_one_file(record_path))
        except:
            data_list.append(pd.DataFrame())
    return data_list



import cv2
import numpy as np


def noise_detection(video_path):
    """
      Get how much of the video that contains video artifacts

      This function returns the ratio of the video that contains artifacts (motion blur, too grainy, static)

      Parameters
      -----------
      video_path : path to a video

      Returns
      -------
      float
          0.284
    """
    cap = cv2.VideoCapture(video_path)
    vid_noise = []
    while cap.isOpened():
        success, image = cap.read()
        if not success:
            break
        image = cv2.cvtColor(image, cv2.COLOR_BGR2RGB)
        s = cv2.calcHist([image], [1], None, [256], [0, 256])
        p = 0.05
        s_perc = np.sum(s[int(p * 255):-1]) / np.prod(image.shape[0:2])
        s_thr = 0.5
        if s_perc > s_thr:
            vid_noise.append(1)
        else:
            vid_noise.append(0)
    return round((sum(vid_noise)/len(vid_noise)), 3)



import math

import cv2

from PIL import Image, ImageStat


def brightness_frame(im):
    """
    Get the brightness of the video frame

    This function returns the brightness of the video frame

    Parameters
    -----------
    im : frame of the video in the form of an image

    Returns
    -------
    float
        28.32078
    """

    im = Image.fromarray(im)
    stat = ImageStat.Stat(im)
    r,g,b = stat.mean

    return math.sqrt(0.241*(r**2) + 0.691*(g**2) + 0.068*(b**2))


def video_brightness(video_file):
    """
    Get the average brightness of the video file

    This function returns the average brightness of the frames of a video file

    Parameters
    -----------
    video_file : path to a specific video 

    Returns
    -------
    float
        125.7859249810
    """

    cap = cv2.VideoCapture(video_file)
    bright = list()
    while(True):
        ret, frame = cap.read()
        if ret: 
            bright.append(brightness_frame(frame))
        else:
            return round(sum(bright)/len(bright), 3) if len(bright) > 0 else 0



from scipy.io import wavfile


def sample_rate(wav_path):
    """
    Get the sample rate of an audio

    This function returns the sample rate of a .wav audio

    Parameters
    -----------
    wav_path : the path of a .wav audio

    Returns
    -------
    int: sample rate of the audio
        "16000"
    """
    sample_rate, wav_data = wavfile.read(wav_path, 'rb')
    return int(sample_rate)



import ast
import os
import subprocess


def creation_time(video_path):
    """
    Get the creation time of a video

    This function returns the date of when the video was created

    Parameters
    -----------
    video_path : path to a specific video 

    Returns
    -------
    string
        "2022-05-29 23:22:59.599607"

    """
    byteD = subprocess.check_output(['ffprobe', '-v', 'quiet', video_path, '-print_format',
                                     'json', '-show_entries', 'stream=index,codec_type:stream_tags=creation_time:format_tags=creation_time'])
    dict = byteD.decode("UTF-8")
    mydata = ast.literal_eval(dict)
    try:
        time = mydata['format']['tags']['creation_time']
    except(KeyError):
        time = 'N/A'
    return str(time).split('.')[0]



import cv2


def fps(video_path):
    """
    Get the frame rate of the video file

    This function returns the frame rate (fps) of the video file

    Parameters
    -----------
    video_path : path to the specific video

    Returns
    -------
    int
        29
    """

    video_data = cv2.VideoCapture(video_path)
    fps_vid = int(video_data.get(cv2.CAP_PROP_FPS))
    return fps_vid


import os 


def audio_paths(audio_dir):
    """
    Get the paths of the audio files

    This function stores the paths of all audio files in a list

    Parameters
    -----------
    audio_dir : directory containing all the audio diaries

    Returns
    -------
    list
        ['/path/to/audio_file1', '/path/to/audio_file2', ... '/path/to/audio_filen']
    """

    audio_paths = []
    for filename in os.listdir(audio_dir):
        if not filename.startswith('.'):
            audio_paths.append(str(os.path.join(audio_dir, filename)))

    return audio_paths


from scipy.io import wavfile
import sys
import numpy as np


def to_array(wav_path, print_full=False):
    """
    Convert an audio file to an array that can be used for plotting graphs and further audio exploration. It can be printed in full, but not recommended since it can be a very large array.

    This function takes the path of a wav audio file and returns the audio array

    Parameters
    -----------
    wav_path : path of a .wav audio
    print_full (Optional, default = False): If True, the full non-truncated representation of the array is printed

    Returns
    -------
    arr
        numpy array of the audio
        "[ 0  0  0 ...  8  0 -2]"
    """
    rate, data = wavfile.read(wav_path, 'rb')
    if print_full == True:
        np.set_printoptions(threshold=sys.maxsize)
    arr = np.array(data)
    return arr



import numpy as np
from scipy import *
from QPrism.Sensor.pipeline_functions.preprocessing import *

# def Signal_to_Noise(x):
#   n = len(x)
#   t = range(n)
#   signal = x
#   orders = range(5,40)
#   sse1 = np.zeros(len(orders))
#   np.savetxt('x.txt', x)
#   for ri in range(len(orders)):
#     yHat = np.polyval(polyfit(t,signal,orders[ri]),t)
#     sse1[ri] = np.sum( (yHat-signal)**2 )/n
#   # Bayes information criterion
#   bic = n*np.log(sse1) + orders*np.log(n)
#   # best parameter has lowest BIC
#   bestP = min(bic)
#   idx = np.argmin(bic)

#   ## now repeat filter for best (smallest) BIC
#   # polynomial fit
#   polycoefs = polyfit(t,signal,orders[idx])
#   # estimated data based on the coefficients
#   yHat = polyval(polycoefs,t)
#   # filtered signal is residual
#   filtsig = signal - yHat
#   SNR = 10*np.log10(np.mean(abs(filtsig))/np.mean(abs(yHat)))
#   return SNR
  
  
def Signal_to_Noise(arr):
    arr = arr[~np.isnan(arr)]
    arr = arr - np.mean(arr)
    arr = np.asarray(arr)
    abs = np.abs(arr)
    mean = abs.mean(axis=0)
    std_dev = arr.std(axis=0, ddof=1)
    snr = np.where(std_dev == 0, 0, mean/std_dev)
    return 20 * np.log10(snr)


def snr(record, feature_names):
    record = features_to_float(record, feature_names)
    #feat_values = [Signal_to_Noise([r[feature] for r in record]) for feature in feature_names]
    feat_values = []
    for feature in feature_names:
        feat_values.append(Signal_to_Noise(record[feature]))
    return np.mean([f for f in feat_values if not math.isnan(f)])


import cv2 

from QPrism.Video.frame_rate import fps


def video_length(video_path):
    """
    Get the length of the video file

    This function returns the video length (seconds) of the video file

    Parameters
    -----------
    video_path : path to the specific video

    Returns
    -------
    int
        101
    """

    video_data = cv2.VideoCapture(video_path)
    frames = video_data.get(cv2.CAP_PROP_FRAME_COUNT)
    video_length = int(frames / fps(video_path))
    return video_length


from pathlib import Path
from os import path
from pydub import AudioSegment
import os


def to_mono_wav(aud_path):
    """
    Convert mp3 files to mono wav files

    This function takes in mp3 audio files and converts them into mono wav files to be used for deep learning and further exploration.

    Parameters
    -----------
    mp3_path : path to the .mp3 file

    Returns
    -------
    NONE
        Saves the .wav file in the same working directory
        "audio_Wav.wav"
    """
    split_tup = os.path.splitext(aud_path)
    format = (split_tup[-1])

    if str(format) == '.wav':
        sound = AudioSegment.from_wav(aud_path)
        sound = sound.set_channels(1)
        audio_name = str(Path(aud_path).stem)
        wav_name = Path(audio_name + "_Mono.wav")
        sound.export(wav_name, format="wav")
    elif str(format) == '.mp3':
        sound = AudioSegment.from_mp3(aud_path)
        sound = sound.set_channels(1)
        audio_name = str(Path(aud_path).stem)
        wav_name = Path(audio_name + "_Wav.wav")
        sound.export(wav_name, format="wav")
    else:
        "Please enter an .mp3 or .wav file"



import pandas as pd
import numpy as np
from QPrism.Sensor.data_processing import *
from QPrism.Sensor.pipeline_functions.signal_to_noise import *


def compute_SNR_single(record_df):
    feature_name = get_feature_name_record_df(record_df)
    SNR = snr(record_df, feature_names=feature_name)
    return SNR


def compute_SNR_multiple(data_list):
    SNR_list = []
    for record in data_list:
        record_SNR = compute_SNR_single(record)
        SNR_list.append(record_SNR)
    return np.mean(SNR_list)
    


import pandas as pd
import numpy as np
from QPrism.Sensor.data_processing import *
from collections import Counter

def compute_SCR_multiple(data_list):
    channel_num_list = []
    incomplete_record_count = 0
    for df in data_list:
        channel_num_list.append(df.shape[1]-1)
    mode = Counter(channel_num_list).most_common(1)[0][0]
    for channel_num in channel_num_list:
        if channel_num != mode:
            incomplete_record_count += 1
    SCR = (len(channel_num_list)-incomplete_record_count)/len(channel_num_list)
    return SCR


from scipy.io import wavfile
import numpy as np
import pandas as pd
import wave
from pathlib import Path


def to_timeseries_dataframe(wav_path, save=False):
    """
    Get the timeseries data frame of an audio

    This function returns a .csv data frame containing the time stamps and data of each channel in each column

    Parameters
    -----------
    wav_path : path to a .wav audio
    save (optional, default=False): If True saves the data frame as a csv file in the working directory.

    Returns
    -------
    Data Frame: contains the time stamps in the first column and channels with corresponding data in the remaining columns. File is saved in working dorectory.
    """
    raw = wave.open(wav_path)
    nch = raw.getnchannels()
    samplerate, data = wavfile.read(wav_path)
    data_per_channel = [data[offset::nch] for offset in range(nch)]
    duration = data.shape[0]/samplerate
    time = np.arange(0, duration, 1/samplerate)
    aud_data = {'Time': time}
    for i in range(0, nch):
        ch_name = 'Channel_' + str(i+1)
        ch_arr = np.ravel(data_per_channel[i])
        aud_data[ch_name] = np.resize(ch_arr, data.shape[0])
    audio_data = pd.DataFrame(aud_data)
    if save == True:
        audio_name = str(Path(wav_path).stem)
        audio_data.to_csv(audio_name+'.csv')
    return audio_data



import pandas as pd
import numpy as np
from QPrism.Sensor.data_processing import *
from QPrism.Sensor.pipeline_functions.preprocessing import *
from QPrism.Sensor.pipeline_functions.single_function_DQMs import *


def compute_MDR_single(record_df):
    sample_consistence, sample_rate_median, sample_mode = sampling_rate_consistency(record_df)
    MDR = count_missing_segments(record_df, sample_mode)
    return MDR


def compute_MDR_multiple(data_list):
    MDR_list = []
    for record in data_list:
        record_MDR = compute_MDR_single(record)
        MDR_list.append(record_MDR)
    return np.mean(MDR_list)



import pandas as pd
import numpy as np
from QPrism.Sensor.pipeline_functions.single_function_DQMs import *
from QPrism.Sensor.data_processing import *


def compute_VRC_multiple_file(data_list):
    min_max = []
    for record_df in data_list:
        feature_name = get_feature_name_record_df(record_df)
        min_max.append(minmax(record_df, feature_name))
    return consistency(min_max)


import os


def video_paths(video_dir):
    """
    Get the paths of the video files

    This function stores the paths of all video files in a list

    Parameters
    -----------
    video_dir : directory containing all the video diaries

    Returns
    -------
    list
        ['/path/to/video_file1', '/path/to/video_file2', ... '/path/to/video_filen']
    """

    video_paths = []
    for filename in os.listdir(video_dir):
        if not filename.startswith('.'):
            video_paths.append(os.path.join(video_dir, filename))
    return video_paths


import csv
import os

import pandas as pd
from pathlib import Path

# Audio functions
from QPrism.Audio.audio_classification import audio_classification
from QPrism.Audio.audio_length import audio_length
from QPrism.Audio.audio_sample_rate import sample_rate
from QPrism.Audio.audio_rms import rms

# Helpers
from QPrism.Audio.helpers.audio_to_array import to_array
from QPrism.Audio.helpers.audio_paths import audio_paths
from QPrism.Audio.helpers.to_mono_wav import to_mono_wav


class Audio_DQM:
    def __init__(self):
        self.aud_metrics = ['classification', 'audio_length', 'sample_rate', 'rms'
                            'signal_to_noise']

    def observation_duration(self, path: str):
        """
        Get the length of the audio

        This functions calls the audio_length function and 
        gives the length for both a folder of audios or a single audio file.

        Returns the length of the input audio if the input path is a file. 
        Returns a dict matching the input audio files to their length if the input path is a folder.

        Parameters
        ----------
            path : str
                A path to a folder or a file.

        """

        if os.path.isdir(path):
            audio_files = audio_paths(path)
            length = {}

            for file in audio_files:
                if not file.startswith('.'):
                    length[str(Path(file).stem)] = audio_length(file)

            return pd.DataFrame(length.items(), columns=['Audios', 'Length'])

        else:
            return audio_length(path)

    def sampling_rate(self, path: str):
        """
        Get the sample rate of the audio

        This functions calls the sample_rate function and 
        gives the bit rate for both a folder of audios or a single audio file.

        Returns the sample rate of the input audio if the input path is a file. 
        Returns a dict matching the input audio files to their sample rate if the input path is a folder.

        Parameters
        ----------
            path : str
                A path to a folder or a file.
        """

        if os.path.isdir(path):
            audio_files = audio_paths(path)
            s_rate = {}

            for file in audio_files:
                if not file.startswith('.'):
                    s_rate[str(Path(file).stem)] = sample_rate(file)

            return pd.DataFrame(s_rate.items(), columns=['Audios', 'Sample Rate'])

        else:
            return sample_rate(path)

    def voice_classification(self, path: str):
        """
        Get a list of all the sounds in an audio

        This function identifies all the prominent sounds inside the audio

        Returns a list of prominent sounds in the input audio if the input path is a file. 
        Returns a dict matching the input audio files to their list of prominent sounds if the input path is a folder.

        Parameters
        ----------
            path : str
                A path to a folder or a file.

        """
        if os.path.isdir(path):
            audio_files = audio_paths(path)
            sounds = {}

            for file in audio_files:
                if not file.startswith('.'):
                    sounds[str(Path(file).stem)] = audio_classification(file)

            return pd.DataFrame(sounds.items(), columns=['Audios', 'Voices'])

        else:
            return audio_classification(path)

    def RMS(self, path: str):
        """
        Compute the RMS value of the audio. This can be used to distinguish audios that are louder from each other.

        Returns the root mean squared value of the audio file(s).
        Returns a dict matching the input audio files to their RMS value if the input path is a folder.
       
        Parameters
        -----------
            path : str
                A path to a folder or a file.
        """

        if os.path.isdir(path):
            audio_files = audio_paths(path)
            rms_values = {}

            for file in audio_files:
                if not file.startswith('.'):
                    rms_values[str(Path(file).stem)] = rms(file)

            return pd.DataFrame(rms_values.items(), columns=['Audios', 'RMS'])

        else:
            return rms(path)

    def save_csv(self, path: str, output_path: str):
        """
        Creates a csv file consists of all the audio metrics.

        Saves the csv file to output_path.

        Parameters
        ----------
            path : str
                A path to single audio file or folder
            output_path : str
                path to the output csv file

        """

        if os.path.isdir(path):
            audio_files = audio_paths(path)
        else:
            audio_files = [os.path.abspath(path)]

        with open('audio_metrics.csv', 'w') as file:
            header = ['Audio Name', 'Observation Duration',
                      'Sampling Rate', 'Voice Classification', 'Energy RMS']
            writer = csv.writer(file)
            writer.writerow(header)

        csv_path = output_path
        metrics = pd.DataFrame()

        for i, audio in enumerate(audio_files):
            data_row = []

            data_row.append(str(Path(audio).stem))
            data_row.append(self.observation_duration(audio))
            data_row.append(self.sampling_rate(audio))
            data_row.append(self.voice_classification(audio))
            data_row.append(self.RMS(audio))

            data_series = pd.Series(data_row, index=header)
            metrics = metrics.append(data_series, ignore_index=True)

        print('Metrics saved to ', csv_path)
        metrics.to_csv(csv_path, index=False)



from DQM import DQM_multiple_file, DQM_single_file
from load_data import *
import pandas as pd


def load_sensor(path):
    records = os.listdir(path)
    filename_list = []
    MDR_list = []
    for record in records:
        record_path = path+'/'+record
        filename_list.append(record)
        df = load_data_json_single_file(record_path)
        single_file_DQM = DQM_single_file()
        single_file_DQM.set_APD(False)
        single_file_DQM.set_SCR(False)
        single_file_DQM.set_SNR(False)
        single_file_DQM.set_SRC(False)
        single_file_DQM.set_input_data(df)
        single_file_DQM.compute_DQM()
        MDR_list.append(single_file_DQM.get_MDR())
    return MDR_list, filename_list


def load_user(path):
    sensors = os.listdir(path)
    MDR_sensor_list = []
    filename_sensor_list = []
    for sensor in sensors:
        sensor_path = path+'/'+sensor
        MDR_list, filename_list = load_sensor(sensor_path)
        MDR_sensor_list.append(MDR_list)
        filename_sensor_list.append(filename_list)
    return filename_sensor_list, MDR_sensor_list


if __name__ == '__main__':
    print(load_user('/home/lin/Documents/CAMH/SenseActivity/data/Test/SUBJ00005'))



import pandas as pd
import time
from QPrism.Sensor.load_data import *
import numpy as np
from QPrism.Sensor.data_processing import *
from QPrism.Sensor.pipeline_functions.preprocessing import features_to_float


def std_arr(arr):
    """
    Return the standard deviation for given array
    """
    arr = arr - np.mean(arr)
    arr = np.asarray(arr)
    std_dev = arr.std(axis=0, ddof=1)
    return std_dev


def get_df_feature_name(df):
    """
    Input:
        df: A dataframe
    Return:
        keys: A list of str represents the feature name of given sensor
    """
    # return -1 if the sensor has no record
    record_df = df
    keys = list(record_df.keys())
    keys.remove(keys[0])
    return keys


def compute_IRLR_single(record_df):
    if (record_df.shape[0]<2):
        return 0
    elif (float(record_df.iloc[[len(record_df)-1]][(list(record_df.keys()))[0]])-float(record_df.iloc[[0]][list(record_df.keys())[0]]))<=0:
        return 0
    else:
        feature_names = get_df_feature_name(record_df)
        record = features_to_float(record_df, feature_names)
        std_values = []
        for feature in feature_names:
            std_values.append(std_arr(record[feature].values))
        for std_value in std_values:
            if (std_value==0):
                return 0
    return 1


def compute_IRLR_multiple(data_list):
    record_count = 0
    insufficient_length_count = 0
    delete_list = []
    for i in range (len(data_list)):
        record_df = data_list[i]
        if (len(record_df)==0):
            IRLR_single = 0
        else:
            IRLR_single = compute_IRLR_single(record_df)
        if (IRLR_single==0):
            insufficient_length_count += 1
            delete_list.append(i)
        record_count += 1
    j = 0
    # delete all the insufficient records
    for index in delete_list:
        del data_list[index-j]
        j += 1
    return (record_count-insufficient_length_count)/record_count





import tensorflow as tf
import tensorflow_hub as hub
import numpy as np
import csv
from scipy import signal
from scipy.io import wavfile


def audio_classification(wav_path):
    """
    Get a list of all the sounds in an audio

    This function returns a list of all the prominent sounds inside the audio

    Parameters
    -----------
    wav_path : path to a .wav audio

    Returns
    -------
    list: all sounds in an audio
        ['Speech', 'Whistling', 'Alarm']
    """

    model = hub.load('https://tfhub.dev/google/yamnet/1')

    def class_names_from_csv(class_map_csv_text):
        """Returns list of class names corresponding to score vector."""
        class_names = []
        with tf.io.gfile.GFile(class_map_csv_text) as csvfile:
            reader = csv.DictReader(csvfile)
            for row in reader:
                class_names.append(row['display_name'])

        return class_names

    class_map_path = model.class_map_path().numpy()
    class_names = class_names_from_csv(class_map_path)

    def ensure_sample_rate(original_sample_rate, waveform,
                           desired_sample_rate=16000):
        """Resample waveform if required."""
        if original_sample_rate != desired_sample_rate:
            desired_length = int(round(float(len(waveform)) /
                                       original_sample_rate * desired_sample_rate))
            waveform = signal.resample(waveform, desired_length)
        return desired_sample_rate, waveform

    sample_rate, wav_data = wavfile.read(wav_path, 'rb')
    sample_rate, wav_data = ensure_sample_rate(sample_rate, wav_data)

    waveform = wav_data / tf.int16.max

    scores, embeddings, spectrogram = model(waveform)

    scores_np = scores.numpy()
    probs = scores_np.mean(axis=0)
    ind = np.where(probs >= 0.1)
    infered_class = []
    for indi in ind[0]:
        infered_class.append(class_names[indi])
    return infered_class



import numpy as np
from sklearn.preprocessing import StandardScaler, MinMaxScaler

def standard_scaler(col, feature_names):
    scaler = StandardScaler()
    try:
        return scaler.fit_transform(X=col.to_numpy().reshape(-1, 1))
    except:
        for key in list(feature_names):
            new_vals = scaler.fit_transform(X= np.asarray([c[key] for c in col]).reshape(-1, 1))
            for i, val in enumerate(new_vals):
                col[i][key] = val[0]
        return col


def normalize_timestamp(record, time_keyname):
    try:
        record[time_keyname] = record[time_keyname] - int(record[time_keyname].iloc[0])
        return record
    except:
        start = record[0][time_keyname]
        for i in range(len(record)):
            record[i][time_keyname] =  record[i][time_keyname]- start
        return record


def min_max_scaler(col):
    scaler = MinMaxScaler()
    return scaler.fit_transform(X=col.to_numpy().reshape(-1, 1))


def multi_var_to_uni(record, cols):
    record['univariate_value'] = np.sqrt(np.square(record[cols]).sum(axis=1))
    return record


def features_to_float(record, feature_names):
    try:
        for feature in feature_names:
            record[feature] = record[feature].astype(float)
        return record
    except:
        for i in range(len(record)):
            for feature in feature_names:
                record[i][feature] = float(record[i][feature])
        return record


from QPrism.Sensor.load_data import *
from QPrism.Sensor.APD import compute_APD_multiple, compute_APD_single
from QPrism.Sensor.IRLR import compute_IRLR_multiple, compute_IRLR_single
from QPrism.Sensor.MDR import compute_MDR_multiple, compute_MDR_single
from QPrism.Sensor.RLC import compute_RLC_multiple_file
from QPrism.Sensor.SCR import compute_SCR_multiple
from QPrism.Sensor.SNR import compute_SNR_multiple, compute_SNR_single
from QPrism.Sensor.SRC import compute_SRC_multiple, compute_SRC_single
from QPrism.Sensor.VRC import compute_VRC_multiple_file
from QPrism.Sensor.VDR import compute_VDR_single, compute_VDR_multiple
import QPrism.Sensor.APD as APD
import csv
import time
import numpy as np


class DQM_single_record:
    """
    A class represents the data quality matrix.
    """
    def __init__(self):
        self.output_path = None
        self.score = []
        self.data_list_pd = None
        self.DQM_config = {'SNR': True, 'VDR': True,
                           'SRC': True, 'MDR': True, 'APD': True}
        self.DQM_function = {'SNR': compute_SNR_single, 'VDR': compute_VDR_single,
                             'SRC': compute_SRC_single,
                             'MDR': compute_MDR_single, 'APD': compute_APD_single}
        self.fields = ['IRLR']
        self.stime = None
        self.loading_etime = None
        self.computing_etime = None
    
    def set_input_data(self, df:pd.DataFrame):
        """
        Set the path of input record file.

        Parameters
        ----------
        path : str
            Path to the input record file
        """
        self.data_list_pd = df
    
    def set_SNR(self, included:bool):
        """
        Set whether SNR is included as a metric

        Parameters
        ----------
        included : bool
            A bool value to indicate whether SNR is included in the result DQM
        """
        self.DQM_config['SNR'] = included
    
    def set_VDR(self, included:bool):
        """
        Set whether VDR is included as a metric

        Parameters
        ----------
        included : bool
            A bool value to indicate whether VDR is included in the result DQM
        """
        self.DQM_config['VDR'] = included
    
    def set_SRC(self, included:bool):
        """
        Set whether SRC is included as a metric

        Parameters
        ----------
        included : bool
            A bool value to indicate whether SRC is included in the result DQM
        """
        self.DQM_config['SRC'] = included
    
    def set_MDR(self, included:bool):
        """
        Set whether MDR is included as a metric

        Parameters
        ----------
        included : bool
            A bool value to indicate whether MDR is included in the result DQM
        """
        self.DQM_config['MDR'] = included
    
    def set_APD(self, included:bool):
        """
        Set whether APD is included as a metric

        Parameters
        ----------
        included : bool
            A bool value to indicate whether APD is included in the result DQM
        """
        self.DQM_config['APD'] = included
    
    def compute_DQM(self):
        """
        Compute the DQM for input record under current metrics configuration.
        Must call set_input_path before calling this method.
        """
        self.score = []
        self.fields = ['IRLR']
        self.stime = time.time()
        # You can remove the comment below to print the computation time. Note this will cause compute_individual_DQM in 
        # the DQM_multi_file class to print this info for each record.
        #print("Start computing the DQM... This may take a long time if the dataset is large")
        for key in self.DQM_config.keys():
                if self.DQM_config[key] is True:
                    self.fields.append(key)
        self.score.append(str(compute_IRLR_single(self.data_list_pd)))
        if (float(self.score[-1])!=0):
            for key in self.DQM_config.keys():
                if self.DQM_config[key] is True:
                    self.score.append(str(self.DQM_function[key](self.data_list_pd)))
        else:
            for key in self.DQM_config.keys():
                if self.DQM_config[key] is True:
                    self.score.append(np.NAN)
        self.etime = time.time()
        #print("The total time for computing the DQM is: " + str(self.etime-self.stime) + ' seconds.')
    
    def get_DQM(self):
        """
        Return the computed DQM as a list. compute_DQM must be called before this method.
        """
        return self.score
    
    def get_fields(self):
        """
        Return a list represents the current included metrics in DQM.
        compute_DQM must be called before this method.
        """
        return self.fields

    def get_IRLR(self):
        """
        Return the IRLR score for given input data as str.
        """
        IRLR_index = self.fields.index('IRLR')
        return self.score[IRLR_index]
    
    def get_SNR(self):
        """
        Return the SNR score for given input data as a str.
        SNR must be included in the DQM class.
        """
        try:
            SNR_index = self.fields.index('SNR')
        except ValueError:
            return "SNR is not computed according to the configuration"
        return self.score[SNR_index]
    
    def get_VDR(self):
        """
        Return the VDR score for given input data as a str.
        VDR must be included in the DQM class.
        """
        try:
            VDR_index = self.fields.index('VDR')
        except ValueError:
            return "VDR is not computed according to the configuration"
        return self.score[VDR_index]
    
    def get_SRC(self):
        """
        Return the SRC score for given input data as a str.
        SRC must be included in the DQM class.
        """
        try:
            SRC_index = self.fields.index('SRC')
        except ValueError:
            return "SRC is not computed according to the configuration"
        return self.score[SRC_index]
    
    def get_MDR(self):
        """
        Return the MDR score for given input data as a str.
        MDR must be included in the DQM class.
        """
        try:
            MDR_index = self.fields.index('MDR')
        except ValueError:
            return "MDR is not computed according to the configuration"
        return self.score[MDR_index]
    
    def get_APD(self):
        """
        Return the APD score for given input data as a str.
        APD must be included in the DQM class.
        """
        try:
            APD_index = self.fields.index('APD')
        except ValueError:
            return "APD is not computed according to the configuration"
        return self.score[APD_index]
    
    def get_anomaly_index(self):
        """
        Return the index (row number) of the anamoly points for the input record.
        APD must be included in the DQM class.
        """ 
        return APD.outlier_index
    
    def save_to_file(self, path:str):
        """
        Save the computed DQM as a csv file to the given output path.
        Must call compute_DQM before calling this.

        Parameters
        ----------
        path : str
            Path to the output file
        
        """
        self.output_path = path
        with open(self.output_path, 'w') as result_file:
            write = csv.writer(result_file)
            write.writerow(self.fields)
            write.writerow(self.score)
        print("Data successfuly saved.")


class DQM_multiple_record:
    """
    A class represents the data quality matrix for multiple files in one directory.
    """
    def __init__(self):
        self.output_path = None
        self.score = []
        self.individual_score = []
        self.data_list = None
        self.DQM_config = {'SNR': True, 'SCR': True, 'RLC': True, 'VRC': True,
                           'SRC': True, 'MDR': True, 'APD': True, 'VDR': True}
        self.DQM_function = {'SNR': compute_SNR_multiple, 'RLC': compute_RLC_multiple_file,
                             'SCR': compute_SCR_multiple, 'SRC': compute_SRC_multiple, 'VRC': compute_VRC_multiple_file, 
                             'MDR': compute_MDR_multiple, 'APD': compute_APD_multiple, 'VDR': compute_VDR_multiple}
        self.fields = ['IRLR']
        self.individual_fields = ['IRLR']
        self.stime = None
        self.loading_etime = None
        self.computing_etime = None
    
    def set_input_data(self, df_list):
        """
        Set the input data list of input file.

        Parameters
        ----------
        df_list : list of pd.DataFrame
            A list of dataframe, each dataframe represents a single record
        """
        self.data_list = df_list
    
    def set_RLC(self, included:bool):
        """
        Set whether RLC is included as a metric

        Parameters
        ----------
        included : bool
            A bool value to indicate whether RLC is included in the result DQM
        """
        self.DQM_config['RLC'] = included
    
    def set_SNR(self, included:bool):
        """
        Set whether SNR is included as a metric

        Parameters
        ----------
        included : bool
            A bool value to indicate whether SNR is included in the result DQM
        """
        self.DQM_config['SNR'] = included
    
    def set_VDR(self, included:bool):
        """
        Set whether VDR is included as a metric

        Parameters
        ----------
        included : bool
            A bool value to indicate whether VDR is included in the result DQM
        """
        self.DQM_config['VDR'] = included
    
    def set_VRC(self, included:bool):
        """
        Set whether VRC is included as a metric

        Parameters
        ----------
        included : bool
            A bool value to indicate whether VRC is included in the result DQM
        """
        self.DQM_config['VRC'] = included
    
    def set_SCR(self, included:bool):
        """
        Set whether SCR is included as a metric

        Parameters
        ----------
        included : bool
            A bool value to indicate whether SCR is included in the result DQM
        """
        self.DQM_config['SCR'] = included
    
    def set_SRC(self, included:bool):
        """
        Set whether SRC is included as a metric

        Parameters
        ----------
        included : bool
            A bool value to indicate whether SRC is included in the result DQM
        """
        self.DQM_config['SRC'] = included
    
    def set_MDR(self, included:bool):
        """
        Set whether MDR is included as a metric

        Parameters
        ----------
        included : bool
            A bool value to indicate whether MDR is included in the result DQM
        """
        self.DQM_config['MDR'] = included
    
    def set_APD(self, included:bool):
        """
        Set whether APD is included as a metric

        Parameters
        ----------
        included : bool
            A bool value to indicate whether APD is included in the result DQM
        """
        self.DQM_config['APD'] = included
    
    def compute_avg_DQM(self):
        """
        Compute a DQM for the input records under current metrics configuration.
        The DQM averaged all the metrics that are supported in DQM_single_file.
        In addition to above metrics, it also computes all the metrics only support DQM_multi_file.
        The computed DQM consists only one row. If you wish to get the DQM for each individual record,
        call the compute_individual_DQM method instead.
        Must call set_input_path before calling this method.
        """
        self.score = []
        self.fields = ['IRLR']
        self.stime = time.time()
        print("Start computing the DQM... This may take a long time if the dataset is large")
        for key in self.DQM_config.keys():
                if self.DQM_config[key] is True:
                    self.fields.append(key)
        self.score.append(str(compute_IRLR_multiple(self.data_list)))
        if (float(self.score[-1])!=0):
            for key in self.DQM_config.keys():
                if self.DQM_config[key] is True:
                    self.score.append(str(self.DQM_function[key](self.data_list)))
        else:
            for key in self.DQM_config.keys():
                if self.DQM_config[key] is True:
                    self.score.append(np.NAN)
        self.etime = time.time()
        print("The total time for computing the DQM is: " + str(self.etime-self.stime) + ' seconds.')
    
    def compute_individual_DQM(self):
        """
        Compute a DQM for the input records under current metrics configuration.
        This computes the DQM for each individual record, and store the DQMs as a list.
        If you wish to get the averaged DQM with the metrics only support multiple records,
        call the compute_avg_DQM method instead.
        Must call set_input_path before calling this method.
        """
        self.individual_score = []
        self.individual_fields = ['IRLR']
        self.stime = time.time()
        print("Start computing the DQM... This may take a long time if the dataset is large")
        multi_metrics_only = ['VRC', 'SCR', 'RLC']
        single_metrics = ['SNR', 'VDR', 'SRC', 'MDR', 'APD']
        for key in self.DQM_config.keys():
                if self.DQM_config[key] is True:
                    if key not in multi_metrics_only:
                        self.individual_fields.append(key)
        for record in self.data_list:
            single_DQM = DQM_single_record()
            single_DQM.set_input_data(record)
            for key in single_metrics:
                if key not in self.individual_fields:
                    if key=='SNR':
                        single_DQM.set_SNR(False)
                    elif key=='VDR':
                        single_DQM.set_VDR(False)
                    elif key=='SRC':
                        single_DQM.set_SRC(False)
                    elif key=='MDR':
                        single_DQM.set_MDR(False)
                    elif key=='APD':
                        single_DQM.set_APD(False)
            single_DQM.compute_DQM()
            single_DQM_result = single_DQM.get_DQM()
            self.individual_score.append(single_DQM_result)
        self.etime = time.time()
        print("The total time for computing the DQM is: " + str(self.etime-self.stime) + ' seconds.')
    
    def get_avg_DQM(self):
        """
        Return the computed DQM as a list. compute_DQM must be called before this method.
        """
        return self.score
    
    def get_individual_DQM(self):
        """
        Return the computed individual DQM as a list. Each element in the list represents a single record.
        compute_individual DQM must be called before this method.
        """
        return self.individual_score
    
    def get_avg_fields(self):
        """
        Return a list represents the current included metrics in the multi-file DQM.
        compute_avg_DQM must be called before this method.
        """
        return self.fields
    
    def get_individual_fields(self):
        """
        Return a list represents the current included metrics in the individual file DQM.
        compute_individual_DQM must be called before this method.
        """
        return self.individual_fields

    def get_IRLR(self):
        """
        Return the IRLR score for given input data as a str.
        """
        IRLR_index = self.fields.index('IRLR')
        return self.score[IRLR_index]
    
    def get_SNR(self):
        """
        Return the SNR score for given input data as a str.
        SNR must be included in the DQM class.
        """
        try:
            SNR_index = self.fields.index('SNR')
        except ValueError:
            return "SNR is not computed according to the configuration"
        return self.score[SNR_index]
    
    def get_VDR(self):
        """
        Return the VDR score for given input data as a str.
        VDR must be included in the DQM class.
        """
        try:
            VDR_index = self.fields.index('VDR')
        except ValueError:
            return "VDR is not computed according to the configuration"
        return self.score[VDR_index]
    
    def get_SCR(self):
        """
        Return the SCR score for given input data as a str.
        SCR must be included in the DQM class.
        """
        try:
            SCR_index = self.fields.index('SCR')
        except ValueError:
            return "SCR is not computed according to the configuration"
        return self.score[SCR_index]
    
    def get_SRC(self):
        """
        Return the SRC score for given input data as a str.
        SRC must be included in the DQM class.
        """
        try:
            SRC_index = self.fields.index('SRC')
        except ValueError:
            return "SRC is not computed according to the configuration"
        return self.score[SRC_index]
    
    def get_MDR(self):
        """
        Return the MDR score for given input data as a str.
        MDR must be included in the DQM class.
        """
        try:
            MDR_index = self.fields.index('MDR')
        except ValueError:
            return "MDR is not computed according to the configuration"
        return self.score[MDR_index]
    
    def get_APD(self):
        """
        Return the APD score for given input data as a str.
        APD must be included in the DQM class.
        """
        try:
            APD_index = self.fields.index('APD')
        except ValueError:
            return "APD is not computed according to the configuration"
        return self.score[APD_index]
    
    def get_RLC(self):
        """
        Return the RLC score for given input data as a str.
        RLC must be included in the DQM class.
        """
        try:
            RLC_index = self.fields.index('RLC')
        except ValueError:
            return "RLC is not computed according to the configuration"
        return self.score[RLC_index]
    
    def get_VRC(self):
        """
        Return the VRC score for given input data as a str.
        VRC must be included in the DQM class.
        """
        try:
            VRC_index = self.fields.index('VRC')
        except ValueError:
            return "VRC is not computed according to the configuration"
        return self.score[VRC_index]
    
    def save_avg_DQM_to_file(self, path:str):
        """
        Save the computed DQM as a csv file to the given output path.
        Must call compute_DQM before calling this.

        Parameters
        ----------
        path : str
            Path to the output file
        
        """
        self.output_path = path
        with open(self.output_path, 'w') as result_file:
            write = csv.writer(result_file)
            write.writerow(self.fields)
            write.writerow(self.score)
        print("Data successfuly saved.")
    
    def save_individual_DQM_to_file(self, path:str):
        """
        Save the computed DQM as a csv file to the given output path.
        Must call compute_DQM before calling this.

        Parameters
        ----------
        path : str
            Path to the output file
        
        """
        self.output_path = path
        with open(self.output_path, 'w') as result_file:
            write = csv.writer(result_file)
            write.writerow(self.individual_fields)
            for individual_DQM in self.individual_score:
                write.writerow(individual_DQM)
        print("Data successfuly saved.")


from pyod.models.feature_bagging import FeatureBagging
import pandas as pd
from QPrism.Sensor.pipeline_functions.preprocessing import *

import warnings
warnings.filterwarnings("ignore")

def detect_outliers(record, feature_names):
    if len(feature_names) == 1:
        record[feature_names[0]+'extra42'] = record[feature_names[0]]
        feature_names.append(feature_names[0]+'extra42')
    record = features_to_float(record, feature_names)
    clf = FeatureBagging()
    clf.fit(record[feature_names].fillna(0))
    decision_scores = clf.decision_scores_
    return decision_scores


def outlier_counts(decision_scores):
    std = np.std(decision_scores)
    avg = np.mean(decision_scores)
    return [ i for i, d in enumerate(decision_scores) if d > avg+3*std ]

 
def count_outliers(record, feature_names):
    record = record.copy()
    record = features_to_float(record,feature_names)
    record = pd.DataFrame.from_dict(record)
    record = normalize_timestamp(record, ((list(record.keys()))[0]))
    record = features_to_float(record, feature_names)
    record = multi_var_to_uni(record, feature_names)
    decision_scores = detect_outliers(record, feature_names)

    record['decision_scores'] = decision_scores
    outlier_index = outlier_counts(decision_scores)
    return len(outlier_index)/len(record), outlier_index


def count_anomalous_segments(changepoints, outlier_locations, record_set):
    fail = 0
    succeed = 0
    if len(changepoints) > 0:
        change_ranges = [[0, changepoints[0]]] + [[changepoints[i], changepoints[i + 1]] for i in
                                                  range(len(changepoints[:-1]))] + [[changepoints[-1], len(record_set)]]
    else:
        change_ranges = []
        fail = 1
    for i, change in enumerate(change_ranges):
        if sum([r['univariate_value'] for r in record_set[i]]) != 0:
            for outlier in outlier_locations:
                if outlier > change[1]:
                    succeed += 1
                    break
                if outlier > change[0] and outlier < change[1]:
                    fail += 1
                    break
    return succeed/((succeed+fail) if succeed+fail> 0 else 1)


from QPrism.Sensor.DQM import DQM_multiple_record, DQM_single_record
from QPrism.Sensor.load_data import *
import pandas as pd


if __name__ == '__main__':
    
    # Compute DQM for an arbitrary json record located at data/sensor/json/ in the git repo
    single_file_dqm = DQM_single_record()
    # replace the path to your actual path locally
    df = load_data_json_single_file('path/to/the/json/record')
    single_file_dqm.set_input_data(df)
    single_file_dqm.compute_DQM()
    print(single_file_dqm.get_fields())
    print(single_file_dqm.get_DQM())

    # Compute DQM for an arbitrary csv record located at data/sensor/csv/ in the git repo
    single_csv_file_dqm = DQM_single_record()
    # replace the path to your actual path locally
    df = load_data_csv_one_file('path/to/the/csv/record')
    single_csv_file_dqm.set_input_data(df)
    single_csv_file_dqm.compute_DQM()
    print(single_csv_file_dqm.get_fields())
    print(single_csv_file_dqm.get_DQM())

    # Compute DQM for all the json records located at data/sensor/json in the git repo
    multi_file_dqm = DQM_multiple_record()
    # replace the path to your actual path locally
    df_list = load_data_json_multi_file('path/to/the/json/record/directory')
    multi_file_dqm.set_input_data(df_list)
    multi_file_dqm.compute_avg_DQM()
    print(multi_file_dqm.get_avg_fields())
    print(multi_file_dqm.get_avg_DQM())

    # Compute DQM for all the csv records located at data/sensor/csv in the git repo
    multi_csvfile_dqm = DQM_multiple_record()
    # replace the path to your actual path locally
    df_list = load_data_csv_multi_file('path/to/the/csv/record/directory')
    multi_csvfile_dqm.set_input_data(df_list)
    multi_csvfile_dqm.compute_DQM()
    print(multi_csvfile_dqm.get_fields())
    print(multi_csvfile_dqm.get_DQM())



import urllib.request

def download(url, name):
    name=name+".mp3"
    try:
        print("Downloading starts...\n")
        urllib.request.urlretrieve(url, name)
        print("Download completed..!!")
    except Exception as e:
        print(e)



import os

def video_format(video_path):
    """
    Get the format of the video file

    This function returns a string which is the extension of the video file fed

    Parameters
    -----------
    video_path : path to the specific video

    Returns
    -------
    string
        '.mp4'
    """

    split_tup = os.path.splitext(video_path)
    format = (split_tup[-1])

    return format


from scipy.io import wavfile


def audio_length(wav_path):
    """
    Get the length of an audio

    This function takes the path of a wav audio file and returns its length

    Parameters
    -----------
    wav_path : path of a .wav audio

    Returns
    -------
    float: the length of the audio file in seconds
        "104.5"
    """
    sample_rate, wav_data = wavfile.read(wav_path, 'rb')
    duration = len(wav_data)/sample_rate
    return round(float(duration), 3)



import pandas as pd
import numpy as np
from QPrism.Sensor.data_processing import *
from QPrism.Sensor.pipeline_functions.outlier_detection import count_outliers


def compute_APD_single(record_df):
    feature_name = get_feature_name_record_df(record_df)
    outlier_percent, outlier_rows = count_outliers(record_df, feature_name)
    global outlier_index
    outlier_index = outlier_rows
    return outlier_percent


def compute_APD_multiple(data_list):
    APD_list = []
    for record in data_list:
        record_APD = compute_APD_single(record)
        APD_list.append(record_APD)
    return np.mean(APD_list)



import os
from genericpath import exists
from pathlib import Path
from moviepy.editor import *

def extract_audios(video_path):
    """
    Extract the audio of the video file

    This function extracts the audio of the video file fed and saves in the audio directory

    Parameters
    -----------
    video_path : path to the specific video

    Returns
    -------
    Audio filename (string) 
        'user_1696_video_diary_867227_Audio.mp3'
    """

    video_name = str(Path(video_path).stem)
    audio_dir = 'audio_files'
    os.makedirs(audio_dir, exist_ok=True)
    audio_name = Path(video_name + "_Audio.mp3")

    try:
        if not audio_name.exists():
            videoclip = VideoFileClip(video_path)
            audioclip = videoclip.audio
            audioclip.write_audiofile(os.path.join(audio_dir, audio_name))
            audioclip.close()
            videoclip.close()
        else:
            pass
        return audio_name

    except OSError:
        print('Encountered OSError')
        return None




import pandas as pd
import numpy as np
from QPrism.Sensor.data_processing import *
from QPrism.Sensor.pipeline_functions.single_function_DQMs import valid_data_ratio


def compute_VDR_single(record_df):
    feature_name = get_feature_name_record_df(record_df)
    VDR = valid_data_ratio(record_df, feature_names=feature_name)
    return VDR


def compute_VDR_multiple(data_list):
    VDR_list = []
    for record in data_list:
        record_VDR = compute_VDR_single(record)
        VDR_list.append(record_VDR)
    return np.mean(VDR_list)


import pandas as pd
import numpy as np
from QPrism.Sensor.pipeline_functions.single_function_DQMs import consistency


def compute_RLC(user_data_list):
    """
    Input: 
        user_data_list: A list represents a user's data, which is the same as the list
                        returned by load_data_pd_single in load_data.py.
    Return:
        The RLC for this user
    """
    RLC_sensor = []
    for sensor in user_data_list:
        record_lengths = []
        for record_df in sensor:
            record_lengths.append(len(record_df))
        RLC_sensor.append(consistency(record_lengths))
    return np.mean(RLC_sensor)


def compute_RLC_multiple_file(data_list):
    record_lengths = []
    for record_df in data_list:
        record_lengths.append(len(record_df))
    return consistency(record_lengths)
        